/*
 * Warehouse.java
 * 
 * This is a simple implementation of the Warehouse Web Service.
 * When the checkAvail() operation is called, randomly either an invalid
 * Quote object is returned, or a valid one, with some
 * fixed price and increasing reference number.
 * 
 * NOTE: this is not a real stateful Web Service, Jambition in its current
 * state is not ready to deal with such stateful services. It will be
 * extended to deal with the JAX-WS stateful extensions, next.
 * see: https://jax-ws.dev.java.net/nonav/2.1.3/docs/statefulWebservice.html
 * 
 */
package services;

import java.util.Random;
import javax.jws.Oneway;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;

/**
 *
 * @author Lars Frantzen
 */
@WebService()
public class Warehouse {

    private static int refNumber = 1;
    private Random random = new Random();

    /**
     * Web service operation
     */
    @WebMethod(operationName = "checkAvail")
    public Quote checkAvail(@WebParam(name = "r") QuoteRequest r) {

        Quote q = new Quote();
        q.setProduct(r.getProduct());

        if (random.nextBoolean()) {
            q.setQuantity(r.getQuantity());
            q.setPrice(41.99);
            q.setRefNumber(refNumber++);
        } else {
            q.setQuantity(0);
            q.setPrice(0.);
            q.setRefNumber(0);
        }

        return q;
    }

    /**
     * Web service operation
     */
    @WebMethod(operationName = "cancelTransact")
    @Oneway
    public void cancelTransact(@WebParam(name = "ref") int ref) {
    }

    /**
     * Web service operation
     */
    @WebMethod(operationName = "orderShipment")
    @Oneway
    public void orderShipment(@WebParam(name = "ref") int ref, @WebParam(name = "adr") Address adr) {
    }
}
