/*
 * Jambition.java
 * Copyright (C) 2007,2008 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
package jambition;

import hu.soft4d.minerva.ssm.SSMFactory;
import stsimulator.STSimulator;
import stsimulator.interpreter.javacc.DumontParser;
import stsimulator.sts.Port;
import stsimulator.sts.SymbolicTransitionSystem;
import stsimulator.types.ST_PseudoPosDouble;
import java.awt.Component;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Properties;
import java.util.logging.FileHandler;
import java.util.logging.Logger;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.SwingWorker;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.Service;

/**
 * 
 * @author Lars Frantzen
 * @version Plastic
 */
public class Jambition {

    static Properties defaultProps;
    static Properties props;
    private static Socket solverSocket = null;
    private static Socket qsdeSocket = null;
    private static PrintWriter solverOut;
    private static BufferedReader solverIn;
    private static PrintWriter qsdeOut;
    private static BufferedReader qsdeIn;
    private static Logger logger = Logger.getLogger("jambition");
    private static MainFrame mf;
    private static SymbolicTransitionSystem sts;
    private static String namespace;
    private static STSimulator sim;
    private static URL WSDLUrl;
    private static OnTheFlyTester tester;
    private static TestEventDispatcher ted;
    private static FocusManagerListener fml;
    private static FailureListener fl;
    private static boolean textHandlerAdded = false;
    private static boolean fileHandlerAdded = false;
    private static boolean testingHasStarted = false;
    private static FileHandler theFileHandler;

    static {
        defaultProps = new Properties();
        defaultProps.setProperty("solverPort", "60002");
        defaultProps.setProperty("qsdePort", "60001");
        defaultProps.setProperty("solverAdr", getLocalHost());
        defaultProps.setProperty("qsdeAdr", getLocalHost());
        defaultProps.setProperty("alwaysNewInputs", "false");
        defaultProps.setProperty("useQSDE", "true");
        defaultProps.setProperty("rememberTrace", "false");
        defaultProps.setProperty("DEBUGMODE", "false");
        defaultProps.setProperty("useLogMonitor", "true");
        defaultProps.setProperty("useLogFile", "false");
        defaultProps.setProperty("logFile", "(none)");
        defaultProps.setProperty("numberOfLogfiles", "1");
        defaultProps.setProperty("bytesPerLogFile", "65536");
        defaultProps.setProperty("WSDLUrl", "");
        defaultProps.setProperty("SSMUrl", "");
        defaultProps.setProperty("Service", "");
        defaultProps.setProperty("Port", "");
        defaultProps.setProperty("XMIFile", "");
        defaultProps.setProperty("SSMName", "");
        props = new Properties(defaultProps);
    }

    private static String getLocalHost() {
        try {
            return InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException ex) {
            ex.printStackTrace();
        }
        return "localhost";
    }

    public static void main(String[] args) {

        java.awt.EventQueue.invokeLater(new Runnable() {

            public void run() {
                try {
                    FileInputStream fis = new FileInputStream(System.getProperty("user.home") + System.getProperty("file.separator") + ".jambition.properties");
                    props.load(fis);
                    fis.close();
                } catch (Exception ex) {
                        ;
                }
                mf = new MainFrame();
                mf.setVisible(true);
                fml = new FocusManagerListener(mf.getLocationBar(), mf.getSwitchBar());
                fl = new FailureListener(mf);
            }
        });
    }

    static boolean generateSSMFromXML(URL WSDLUrl, URL SSMUrl, String service, String port, Component parent) {

        if (Boolean.valueOf(props.getProperty("useLogMonitor")) && !textHandlerAdded) {
            logger.addHandler(new JTextAreaHandler(mf.getJTextArea()));
            textHandlerAdded = true;
        }
        if (Boolean.valueOf(props.getProperty("useLogFile")) && !fileHandlerAdded) {
            try {
                theFileHandler = new FileHandler(props.getProperty("logFile"),
                        Integer.valueOf(props.getProperty("bytesPerLogFile")), Integer.valueOf(props.getProperty("numberOfLogfiles")));
                logger.addHandler(theFileHandler);
                fileHandlerAdded = true;
            } catch (SecurityException ex) {
                ex.printStackTrace();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }


        Jambition.WSDLUrl = WSDLUrl;

        try {
            SSMFactory ssmf = SSMFactory.getFactoryFromWSDL(WSDLUrl.toString(), SSMUrl.toString(), service, port);
            sts = ssmf.getSSM();
            namespace = ssmf.getNamespace();
        } catch (Exception ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Couldn't generate SSM (see exception)!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }
        System.out.println("SSM generated:");
        System.out.println(Jambition.sts);
        System.out.println("\n\n");

        try {
            Jambition.sts.attachParsersToSwitches();
        } catch (SymbolicTransitionSystem.NonConsistentSTSException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "The generated SSM is not consistent!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (DumontParser.IllegalTypeException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Illegal type in switches found!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (SymbolicTransitionSystem.SwitchParseException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Switch parse exception!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }
        return true;
    }

    static boolean generateSSMFromUML(String modelName, String ssmName, Component parent) {

        if (Boolean.valueOf(props.getProperty("useLogMonitor")) && !textHandlerAdded) {
            logger.addHandler(new JTextAreaHandler(mf.getJTextArea()));
            textHandlerAdded = true;
        }
        if (Boolean.valueOf(props.getProperty("useLogFile")) && !fileHandlerAdded) {
            try {
                theFileHandler = new FileHandler(props.getProperty("logFile"),
                        Integer.valueOf(props.getProperty("bytesPerLogFile")), Integer.valueOf(props.getProperty("numberOfLogfiles")));
                logger.addHandler(theFileHandler);
                fileHandlerAdded = true;
            } catch (SecurityException ex) {
                ex.printStackTrace();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }


        try {
            SSMFactory ssmf = SSMFactory.getFactoryFromUML(modelName, ssmName);
            sts = ssmf.getSSM();
            namespace = ssmf.getNamespace();
            Jambition.WSDLUrl = new URL(ssmf.getWsdlUri());
        } catch (Exception ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Couldn't generate SSM (see exception)!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }
        System.out.println("SSM generated:");
        System.out.println(Jambition.sts);
        System.out.println("\n\n");

        try {
            Jambition.sts.attachParsersToSwitches();
        } catch (SymbolicTransitionSystem.NonConsistentSTSException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "The generated SSM is not consistent!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (DumontParser.IllegalTypeException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Illegal type in switches found!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (SymbolicTransitionSystem.SwitchParseException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Switch parse exception!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }
        return true;
    }

    public static boolean connectToSolver(Component parent) {
        // Connect To the Prolog Server

        if (solverSocket != null) {
            return true;
        }

        try {
            solverSocket = new Socket(props.getProperty("solverAdr"), Integer.valueOf(props.getProperty("solverPort")));
            //soc.configureBlocking(false);
            logger.info("treeSolver is bound to local port " + solverSocket.getLocalPort());
            solverOut = new PrintWriter(solverSocket.getOutputStream(), true);
            solverIn = new BufferedReader(new InputStreamReader(solverSocket.getInputStream()));
            String welcomeMessage = solverIn.readLine();
            logger.info("treeSolver sends: " + welcomeMessage);
        } catch (UnknownHostException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Unkown treeSolver host!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "IOException with treeSolver!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }
        return true;
    }

    public static boolean connectToQSDE(Component parent) {
        // Connect to the QSDE

        if (qsdeSocket != null) {
            return true;
        }

        try {
            qsdeSocket = new Socket(props.getProperty("qsdeAdr"), Integer.valueOf(props.getProperty("qsdePort")));
            //soc.configureBlocking(false);
            logger.info("qsdeServer is bound to local port " + qsdeSocket.getLocalPort());
            qsdeOut = new PrintWriter(qsdeSocket.getOutputStream(), true);
            qsdeIn = new BufferedReader(new InputStreamReader(qsdeSocket.getInputStream()));
        } catch (UnknownHostException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Unkown Quick Sequence Diagram Editor host!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "IOException with Quick Sequence Diagram Editor!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }
        return true;
    }

    public static void endSocketConnections() {
        if (solverSocket != null) {
            solverOut.println("bye.");
            try {
                solverSocket.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        if (qsdeSocket != null && testingHasStarted) {
            System.out.println(qsdeSocket);
            qsdeOut.println("end");
            try {
                solverSocket.close();
                qsdeSocket.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    public static boolean setupSimulator(Component parent) {
        try {
            sim = new STSimulator(sts, solverSocket, logger);
        } catch (DumontParser.RuntimeViolation ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Simulator discovered a runtime violation!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Simulator had an IOException!", "Error", JOptionPane.ERROR_MESSAGE);
            return false;
        }

        if (Boolean.valueOf(props.getProperty("useQSDE"))) {
            try {
                sim.setDisplaySocket(qsdeSocket, "Jambition");
            } catch (IOException ex) {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(parent, "Could not set the display socket!", "Error", JOptionPane.ERROR_MESSAGE);
                return false;
            }
        }

        sim.setGlobalNewInputs(Boolean.valueOf(props.getProperty("alwaysNewInputs")));
        sim.setRemeberSuspensionTrace(Boolean.valueOf(props.getProperty("rememberTrace")));

        ST_PseudoPosDouble.postPointLength = 2;

        // For Jambition we assume (exactly) one Service and Port being present in the SSM

        System.out.println("WSDLUrl: " + WSDLUrl);
        System.out.println(new QName(namespace, ((stsimulator.sts.Service) sts.getServices().iterator().next()).getName()));
        System.out.println(((Port) ((stsimulator.sts.Service) sts.getServices().iterator().next()).getPorts().iterator().next()).getName());
        System.out.println(namespace);
        
        Service s = Service.create(WSDLUrl, new QName(namespace, ((stsimulator.sts.Service) sts.getServices().iterator().next()).getName()));
        String portname = ((Port) ((stsimulator.sts.Service) sts.getServices().iterator().next()).getPorts().iterator().next()).getName();
        QName portQName = new QName(namespace, portname);

        ted = new TestEventDispatcher(sts, s.createDispatch(portQName, SOAPMessage.class, Service.Mode.MESSAGE), namespace, logger, mf);
        tester = new OnTheFlyTester(parent, sim, ted, logger);
        tester.addPropertyChangeListener(fml);
        tester.addPropertyChangeListener(fl);

        return true;
    }

    static void startTesting(Component parent) {
        if (tester.getState() == SwingWorker.StateValue.DONE) {
            tester = new OnTheFlyTester(parent, sim, ted, logger);
            tester.addPropertyChangeListener(fml);
            tester.addPropertyChangeListener(fl);
        }

        tester.execute();
        testingHasStarted = true;
    }

    static void stopTestingPressed() {
        tester.stopTesting();
    }

    static void endQSDESession() {
        if (qsdeSocket != null) {
            qsdeOut.println("end");
            try {
                qsdeSocket.close();
                qsdeSocket = null;
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    static void testingHasEnded() {
        testingHasStarted = false;
        if (Boolean.valueOf(props.getProperty("useLogFile"))) {
            theFileHandler.close();
        }
        theFileHandler = null;
        fileHandlerAdded = false;
    }

    static boolean simRemembersTrace() {
        return sim.getRememberSuspensionTraces();
    }

    static String getDetailedCoverage() {
        return sim.getCoverageStatus();
    }

    static String getQSDEInput() {
        return sim.suspensionTraceAsQSDE();
    }

    static void saveProperties(Component parent) {
        try {
            FileOutputStream fos = new FileOutputStream(System.getProperty("user.home") + System.getProperty("file.separator") + ".jambition.properties");
            props.store(fos, "The configuration of Jambition");
            fos.close();
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Couldn't save configuration (see exception)!", "Error", JOptionPane.ERROR_MESSAGE);
        }
    }

    static Properties getDefaultProperties() {
        return defaultProps;
    }

    static String getDotInput() {
        return sim.asDotFormat();
    }

    static class FocusManagerListener implements PropertyChangeListener {

        private JProgressBar loc;
        private JProgressBar swi;

        FocusManagerListener(JProgressBar loc, JProgressBar swi) {
            this.loc = loc;
            this.swi = swi;
        }

        public void propertyChange(PropertyChangeEvent e) {
            String propertyName = e.getPropertyName();
            if ("lcov".equals(propertyName)) {
                loc.setValue((int) Math.round(new Double((Double) e.getNewValue())));
            } else if ("scov".equals(propertyName)) {
                swi.setValue((int) Math.round(new Double((Double) e.getNewValue())));
            }

        }
    }

    static class FailureListener implements PropertyChangeListener {

        private MainFrame mf;

        FailureListener(MainFrame mf) {
            this.mf = mf;
        }

        public void propertyChange(PropertyChangeEvent e) {
            String propertyName = e.getPropertyName();
            if ("failure".equals(propertyName)) {
                mf.failureDetected();
            }
        }
    }
}
