/*
 * TestEventDispatcher.java
 * Copyright (C) 2007,2008 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
package jambition;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.logging.Logger;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.ws.Dispatch;
import javax.xml.ws.WebServiceException;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import stsimulator.sts.InstantiatedMessage;
import stsimulator.sts.InteractionVariable;
import stsimulator.sts.Message;
import stsimulator.sts.MessageKind;
import stsimulator.sts.Operation;
import stsimulator.sts.Port;
import stsimulator.sts.SymbolicTransitionSystem;
import stsimulator.sts.Valuation;
import stsimulator.types.ComplexType;
import stsimulator.types.ComplexTypeInstance;
import stsimulator.types.Enumeration;
import stsimulator.types.EnumerationInstance;
import stsimulator.types.ST_Boolean;
import stsimulator.types.ST_BooleanInstance;
import stsimulator.types.ST_Double;
import stsimulator.types.ST_DoubleInstance;
import stsimulator.types.ST_Int;
import stsimulator.types.ST_IntInstance;
import stsimulator.types.ST_PosDouble;
import stsimulator.types.ST_PosDoubleInstance;
import stsimulator.types.ST_PosInt;
import stsimulator.types.ST_PosIntInstance;
import stsimulator.types.ST_PseudoPosDouble;
import stsimulator.types.ST_PseudoPosDoubleInstance;
import stsimulator.types.ST_String;
import stsimulator.types.ST_StringInstance;
import stsimulator.types.Type;
import stsimulator.types.TypeInstance;

/**
 *
 * @author Lars Frantzen
 * @version Plastic
 */
public class TestEventDispatcher {

    private static String mynamespace = "jambition";
    private SymbolicTransitionSystem sts;
    private String namespaceURI;
    private Dispatch<SOAPMessage> myDispatcher;
    private Logger logger;
    private MainFrame mf;

    /** Creates a new instance of TestEventDispatcher */
    public TestEventDispatcher(SymbolicTransitionSystem sts, Dispatch<SOAPMessage> myDispatcher, String namespaceURI, Logger logger, MainFrame mainFrame) {
        this.sts = sts;
        this.logger = logger;
        this.myDispatcher = myDispatcher;
        this.namespaceURI = namespaceURI;
        this.mf = mainFrame;
    }

    void invokeOneWay(InstantiatedMessage im) {
        SOAPMessage message = InstantiatedMessage2SOAPMessage(im);

        if (Boolean.valueOf(Jambition.props.getProperty("DEBUGMODE"))) {
            logger.info("Invoking OneWay SOAP request " + SOAPMessageToString(message));
        }

        try {
            myDispatcher.invokeOneWay(message);
        } catch (WebServiceException exception) {
            logger.severe("Could not invoke RequestResponse SOAP request: " + exception.getMessage());
            mf.externalExceptionThrown(exception.getMessage());
        }
    }

    InstantiatedMessage invokeSync(InstantiatedMessage im) {
        SOAPMessage message = InstantiatedMessage2SOAPMessage(im);

        if (Boolean.valueOf(Jambition.props.getProperty("DEBUGMODE"))) {
            logger.info("Invoking RequestResponse SOAP request " + SOAPMessageToString(message));
        }

        SOAPMessage response = null;
        try {
            response = myDispatcher.invoke(message);
        } catch (WebServiceException exception) {
            logger.severe("Could not invoke RequestResponse SOAP request: " + exception.getMessage());
            mf.externalExceptionThrown(exception.getMessage());
        }

        if (Boolean.valueOf(Jambition.props.getProperty("DEBUGMODE"))) {
            logger.info("Received SOAP response: " + SOAPMessageToString(response));
        }
        // TODO: If the SOAP request is misformed, or an exception it thrown at the SUT side,
        // then this is encoded in the SOAPMessage and
        // has to be checked. No local expection is thrown here
        // (but it happens when using Source Dispatchers, strange...)

        SOAPBody body = null;
        try {
            body = response.getSOAPBody();
        } catch (SOAPException ex) {
            System.out.println("EXCEPTION POINT 4 !!!!!!!!!!!!!!!!!");
            ex.printStackTrace();
        }

        Node methodNode = ((Node) (body.getChildElements().next()));

        Operation op = im.getMessage().getOperation();
        Message returnMessage = null;
        try {
            // returnMessage = findSTSMessage(messageName, MessageKind.OUTPUT, op, im.getMessage().getPort());
            returnMessage = findSTSMessage(MessageKind.OUTPUT, op);
        } catch (Exception ex) {
            ex.printStackTrace();
            System.exit(0);
        }

        ArrayList<InteractionVariable> returnType = returnMessage.getType();
        NodeList parameterNode = methodNode.getChildNodes();
        Valuation returnValuation = new Valuation();
        int i = 0;
        for (Iterator it = returnType.iterator(); it.hasNext();) {
            InteractionVariable var = (InteractionVariable) it.next();
            TypeInstance tin = instanceFromNodeList(var.getType(), parameterNode.item(i++));
            returnValuation.addSingleValuation(var.getName(), tin);
        }

        InstantiatedMessage finalResult = new InstantiatedMessage(returnMessage, returnValuation);

//        if (Boolean.valueOf(Jambition.props.getProperty("DEBUGMODE"))) {
//            logger.info("Assembled resulting instantiated message: " + finalResult);
//        }

        return finalResult;
    }

    private static String instanceAsSOAPSequence(Type type, TypeInstance instance) {
        StringBuffer result = new StringBuffer();

        if (type instanceof ComplexType) {
            ComplexType cti = (ComplexType) type;
            Object[][] fields = cti.getFields();
            Object[] values = (Object[]) instance.getValue();

            for (int i = 0; i < fields.length; i++) {
                result.append('<' + (String) fields[i][0] + '>');
                result.append(instanceAsSOAPSequence((Type) fields[i][1], (TypeInstance) values[i]));
                result.append("</" + (String) fields[i][0] + '>');
            }
        } else if (type instanceof Enumeration) {
            result.append((String) ((EnumerationInstance) instance).getValue());
        } else // for Simple Types the toString() does the job
        {
            result.append(instance.toString());
        }

        return result.toString();
    }

    private SOAPMessage InstantiatedMessage2SOAPMessage(final InstantiatedMessage im) {
        String inputMessage = im.getMessage().getName();
        ArrayList<InteractionVariable> type = im.getMessage().getType();
        Hashtable<String, TypeInstance> valuationTable = im.getValuation().getValuationTable();

        StringBuffer paramString = new StringBuffer();

        for (Iterator it = type.iterator(); it.hasNext();) {
            InteractionVariable iv = (InteractionVariable) it.next();
            paramString.append('<' + iv.getName() + '>');
            paramString.append(instanceAsSOAPSequence(iv.getType(), valuationTable.get(iv.getName())));
            paramString.append("</" + iv.getName() + '>');
        }

        SOAPMessage message = assembleSOAPMessage(inputMessage, paramString.toString());
        return message;
    }

    private SOAPMessage assembleSOAPMessage(String opname, String parameters) {
        String request = "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:" + mynamespace + "=\"" + namespaceURI + "\"" + "><soapenv:Body><" + mynamespace + ":" + opname + ">" +
                parameters + "</" + mynamespace + ":" + opname + "></soapenv:Body></soapenv:Envelope>";
        SOAPMessage message = null;
        try {
            MessageFactory factory = MessageFactory.newInstance();
            message = factory.createMessage();
            message.getSOAPPart().setContent((Source) new StreamSource(new StringReader(request)));
            message.saveChanges();
        } catch (SOAPException e) {
            System.out.println("------------ EXCEPTION assembleSOAPMessage ------------");
            e.printStackTrace();
        }

        return message;
    }

    private static String SOAPMessageToString(final SOAPMessage response) {
        String xmlString = null;
        try {
            Source result = response.getSOAPPart().getContent();
            xmlString = sourceToXMLString(result);
        } catch (SOAPException e) {
            System.out.println("------------ EXCEPTION SOAPMessageToString ------------");
            e.printStackTrace();
        }
        return xmlString;
    }

    private static String sourceToXMLString(Source result) {
        String xmlResult = null;
        try {
            TransformerFactory factory = TransformerFactory.newInstance();
            Transformer transformer = factory.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            OutputStream out = new ByteArrayOutputStream();
            StreamResult streamResult = new StreamResult();
            streamResult.setOutputStream(out);
            transformer.transform(result, streamResult);
            xmlResult = streamResult.getOutputStream().toString();
        } catch (TransformerException e) {
            e.printStackTrace();
        }
        return xmlResult;
    }

    // For single port specifications it is enough to know the MessageKind and the Operation
    private Message findSTSMessage(MessageKind kind, Operation op) throws Exception {
        HashSet<Message> messages = sts.getMessages();

        for (Iterator it = messages.iterator(); it.hasNext();) {
            Message m = (Message) it.next();

            if (m.getKind() != MessageKind.UNOBSERVABLE) {
                if ((m.getKind() == kind) && (m.getOperation().equals(op))) {
                    return m;
                }
            }
        }
        throw new Exception("Cannot find the STS message belonging to the SOAP response!");
    }

    private Message findSTSMessage(String messageName, MessageKind kind, Operation op, Port port) throws Exception {

        HashSet<Message> messages = sts.getMessages();

        for (Iterator it = messages.iterator(); it.hasNext();) {
            Message m = (Message) it.next();

            if (m.getKind() != MessageKind.UNOBSERVABLE) {
                if ((m.getName().equals(messageName)) && (m.getKind() == kind) && (m.getOperation().equals(op)) && (m.getPort().equals(port))) {
                    return m;
                }
            }
        }

        throw new Exception("Cannot find the STS message belonging to the SOAP response!");
    }

    private TypeInstance instanceFromNodeList(Type type, Node node) {

            if (node == null) {
                logger.severe("SOAP response does not contain data for type: " + type);
                mf.externalExceptionThrown("SOAP response does not contain data for type:\n" + type);
            }
        
        TypeInstance result = null;
        if (type instanceof ST_Boolean) {
            result = new ST_BooleanInstance(Boolean.parseBoolean(node.getTextContent()));
        }
        if (type instanceof ST_Double) {
            result = new ST_DoubleInstance(Double.parseDouble(node.getTextContent()));
        }
        if (type instanceof ST_Int) {
            result = new ST_IntInstance(Integer.parseInt(node.getTextContent()));
        }
        if (type instanceof ST_PosDouble) {
            result = new ST_PosDoubleInstance(Double.parseDouble(node.getTextContent()));
        }
        if (type instanceof ST_PosInt) {
            result = new ST_PosIntInstance(Integer.parseInt(node.getTextContent()));
        }
        if (type instanceof ST_PseudoPosDouble) {
            result = new ST_PseudoPosDoubleInstance(Double.parseDouble(node.getTextContent()));
        }
        if (type instanceof ST_String) {
            result = new ST_StringInstance(node.getTextContent());
        }
        if (type instanceof Enumeration) {
            result = new EnumerationInstance((Enumeration) type, node.getTextContent());
        }
        if (type instanceof ComplexType) {
            ComplexType ctype = (ComplexType) type;
            Object[][] fields = ctype.getFields();
            Object[] fieldvalues = new Object[fields.length];

            NodeList childNodes = node.getChildNodes();
            if (childNodes.getLength() != fields.length) {
                logger.severe("SOAP response contains complex type with wrong number of elements, must match: " + ctype);
                mf.externalExceptionThrown("Wrong number of complex type elements in SOAP response, must match:\n" + ctype);
            }

            for (int i = 0; i < fields.length; i++) {
                fieldvalues[i] = instanceFromNodeList((Type) fields[i][1], childNodes.item(i));
            }
            result = new ComplexTypeInstance(fieldvalues);
        }
        return result;
    }
}
